/*
* Politecnico di Milano
* Corso Informatica Medica 2012/2013 - Prof. Santambrogio
* Esercizi di: Alessandro A. Nacci
* E-mail: alessandro.nacci@mail.polimi.it
* Web: www.alessandronacci.com
* ------------------------------------
*/


/****
*
*	FORZA 4
*	---------------------------------	
*
*	Implementare il gioco del Forza 4.
*
*/



#include <stdio.h>
#include <string.h>

// Impostazioni schermo
#define SCREEN_H 8 
#define SCREEN_W 9
#define RISOLUZIONE 72 	//SCREEN_H * SCREEN_W
#define DIAG 13

// Impostazioni memorizzazione
#define MAX_PUNTI_FORMA 64
#define MAX_NUMERO_FORME 10

// Impostazioni sistema
#define LINEE_TERMINALE 50

// Impostazioni FORZA 4
#define MAX_NOME_GIOCATORE 10
#define NUMERO_GIOCATORI 2
#define FORZA 4

/*
===============================================
	DEFINIZIONE DEI TIPI DI DATO
===============================================

	Prestare attenzione!
	---------------------------------
	In questa sezione vengono definiti TIPI DI DATO e >> NON << variabili.
	Il tipo di dato rappresenta solo la 'forma' e la 'dimensione in memoria' di una data variabile.
	Il tipo di dato NON contiene alcun valore.

*/


// Elenco delle forme supportate dal programma
typedef enum {F_PUNTO, F_LINEA,
			  F_POLIGONO_QUADRILATERO, F_GENERICA} categoria_forma;			// Elenco delle forme supportate dal programma
typedef enum {D_VERTICALE, D_ORIZZONTALE} direzione;						// Elenco direzioni possibili per le linee

typedef struct {
	int x;
	int y;
	char valore;
} punto_schermo;																	// TIPO DI DATO per la creazione di variabili di tipo 'punto'

typedef struct {
	punto_schermo pixels[MAX_PUNTI_FORMA];											// Elenco dei pixel che compongono la forma
	int numero_pixel;														// Numero dei pixel che compongono la forma
	categoria_forma categoria;												// Categoria della forma
} forma;																	// TIPO DI DATO per la creazione di variabili di tipo 'forma'


typedef struct {
	char nome[MAX_NOME_GIOCATORE];
	char simbolo_pedina;
} giocatore;

/*
===============================================
	PROTOTIPI DELLE FUNZIONI
===============================================
*/

// Funzioni per la visualizzazione a video
void inizializza_schermo(char schermo[SCREEN_W][SCREEN_H]);						// Riempie di caratteri vuoti la matrice in ingresso 'schermo'
void inserisci_bordi(char schermo[SCREEN_W][SCREEN_H]);							// Inserisce il carattere '%' sui bordi della matrice 'schermo'
void disegna_schermo(char schermo[SCREEN_W][SCREEN_H]);							// Disegna a schermo la matrice in ingresso 'schermo'
void aspetta_invio();															// Quando lanciata, mette il programma in attesa di un INVIO da parte dell'utente
void pulisci_terminale();														// Cancella il contenuto del termianale su cui viene stampato l'output del programma

// Modifica delle matriche che rappresenta lo schemo
void disegna_forma(forma f, punto_schermo p, char schermo[SCREEN_W][SCREEN_H]);			// Disegna una data un generica forma 'f' nella posizione 'p' dello schermo 'schermo'
punto_schermo crea_punto_schermo(int x, int y, char valore);							// Resistuisce una variabile strutturata che rapprensenta il punto (x,y) di valore 'valore'

// Generazione delle forme
forma genera_quadrato(int dim, char carattere);									// Restituisce una forma quadrata di lato 'dim'
forma genera_linea(int dim, direzione direzione_linea, char carattere);			// Restituisce una forma linea di lunghezza 'dim'
forma genera_punto(char carattere);												// Restituisce una forma punto (linea di lunghezza 1)
forma genera_polinomio(int a2, int a1, int a0, int dim, char carattere);		// Restituisce la forma che descrive i primi 'dim' punti di un polinomio nella forma y = a2 * x^2 + a1 * x + a0

// Funzioni Forza 4

// Restituisce una variabile 'giocatore' dato il nome del giocatore e il simbolo della pedina
giocatore crea_giocatore(char nome[MAX_NOME_GIOCATORE], char pedina);				

// Richiede a schermo dove inserire pedina. Restisce '1' se tutto ok, '0' in caso di errore
int richiedi_inserimento_pedina();													

// Inserisce la pedina 'pedina' nella colonna 'colonna' dello campo di gioco 'schermo'
int inserisci_pedina(char schermo[SCREEN_W][SCREEN_H], int colonna, char pedina);	


int controlla_vettore(char* vettore, int dim_vettore, char occorrenza);				// Controlla che in un vettore di siano FORZA occorrenze consecutive di uno stesso simbolo 'occorrenza'
int controlla_orizzontale(char schermo[SCREEN_W][SCREEN_H], char simbolo_pedina);	// Controlla che ci sia una vincita nelle righe orizzontali delle schermo 'schermo' 
int controlla_verticale(char schermo[SCREEN_W][SCREEN_H], char simbolo_pedina);		// Controlla che ci sia una vincita nelle righe verticali delle schermo 'schermo' 
int controlla_obliquo(char schermo[SCREEN_W][SCREEN_H], char simbolo_pedina);		// Controlla che ci sia una vincita nelle righe oblique delle schermo 'schermo' 
int controlla_vincita(char schermo[SCREEN_W][SCREEN_H], char simbolo_pedina);		// Controlla che ci sia una vincinta nell'intero schermo 'schermo'
int estra_controlla_vettore_obliquo(int x, int y, char schermo[SCREEN_W][SCREEN_H], char occorrenza, int orientamento);


/*
===============================================
	MAIN
===============================================
*/


int main(){

	char schermo[SCREEN_W][SCREEN_H];
	giocatore giocatori[NUMERO_GIOCATORI];
	
	int giocatore_corrente = 0;
	int vincitore = -1;
	int colonna_richiesta;

	// Disegnamo un quadrato
	giocatori[0] = crea_giocatore("MARIO", 'o');
	giocatori[1] = crea_giocatore("PAOLO", 'x');

	inizializza_schermo(schermo);
	disegna_schermo(schermo);
	while(vincitore == -1)
	{
		printf ("Giocatore %d \n", giocatore_corrente + 1);
		colonna_richiesta = richiedi_inserimento_pedina();
		if (inserisci_pedina(schermo, colonna_richiesta, giocatori[giocatore_corrente].simbolo_pedina)){
			if (controlla_vincita(schermo, giocatori[giocatore_corrente].simbolo_pedina))
			{
				printf("Il giocatore %d, %s ha vinto!\n\n", giocatore_corrente + 1, giocatori[giocatore_corrente].nome);
				return 0;
			}
			giocatore_corrente = (giocatore_corrente + 1) % NUMERO_GIOCATORI;
		} else {
			printf("La colonna inserita non è valida. Seleziona un'altra colonna.\n");
			aspetta_invio();
			aspetta_invio();	// RISOLVERE QUESTO PROBLEMA
		}
		disegna_schermo(schermo);
	}



}


/*
===============================================
	DEFINIZIONE DELLE FUNZIONI
===============================================
*/

int controlla_vettore(char* vettore, int dim_vettore, char occorrenza)
{


	int i;
	int occorrenze = 0;

	for (i = 1; i < dim_vettore; i++){

		if (vettore[i] == vettore[i-1] && vettore[i] == occorrenza)
		{
			occorrenze++;
			if (occorrenze == FORZA-1)
				return 1;
		}
		else
		{		
			occorrenze = 0;
		}
	}
	return 0;
}

int estra_controlla_vettore_obliquo(int x, int y, char schermo[SCREEN_W][SCREEN_H], 
										char occorrenza, int orientamento)
{


	char vettore[DIAG];
	int curr_x, curr_y;
	int i = 0;

	do {
		curr_x = x + (orientamento) * i;	
		curr_y = y +  i;
		vettore[i] = schermo[curr_x][curr_y];
		i++;

	} while (curr_x >= 0 &&  curr_x < SCREEN_W && curr_y >= 0 && curr_y < SCREEN_H);

	return controlla_vettore(vettore, i-1, occorrenza);

}

int controlla_obliquo(char schermo[SCREEN_W][SCREEN_H], char occorrenza)
{
	
	int x,y,i;
	int risultato_controllo = 0;

	for (y = 0; y < SCREEN_H; y++)
		for (x = 0; x < SCREEN_W; x++)
		{
			risultato_controllo = estra_controlla_vettore_obliquo(x,y,schermo,occorrenza, 1);
			if (risultato_controllo) return 1;

			risultato_controllo = estra_controlla_vettore_obliquo(SCREEN_W-x,y,schermo,occorrenza, -1);
			if (risultato_controllo) return 1;
		}




	return 0;
}

int controlla_verticale(char schermo[SCREEN_W][SCREEN_H],
						char simbolo_pedina)
{

	int x,y;
	char vettore[SCREEN_W];
	int risultato = 0;


	for (x = 0; x< SCREEN_W; x++)
	{
		for (y = SCREEN_H-2; y >= 0; y--)
		{		
			vettore[y] = schermo[x][y];
		}

		risultato = controlla_vettore(vettore, SCREEN_W, simbolo_pedina);
		if (risultato == 1)
			return 1;
	}

	return 0;

}

int controlla_orizzontale(char schermo[SCREEN_W][SCREEN_H], 
							char simbolo_pedina)
{

	int x,y;
	char vettore[SCREEN_W];
	int risultato = 0;

	for (y = SCREEN_H-2; y >= 0; y--)
	{
		for (x = 0; x< SCREEN_W; x++){
			vettore[x] = schermo[x][y];
		}

		risultato = controlla_vettore(vettore, SCREEN_W, simbolo_pedina);
		if (risultato == 1)
			return 1;
	}

	return 0;

}

int controlla_vincita(char schermo[SCREEN_W][SCREEN_H], char simbolo_pedina)
{
	int result = 0;

	result += controlla_orizzontale(schermo, simbolo_pedina);
	result += controlla_verticale(schermo, simbolo_pedina);
	result += controlla_obliquo(schermo, simbolo_pedina);

	return result;
}

int inserisci_pedina(char schermo[SCREEN_W][SCREEN_H], int colonna, char simbolo_pedina)
{
	int y;
	punto_schermo p;
	forma pedina;

	for (y = SCREEN_H-2; y >= 0; y--)
	{

		if (schermo[colonna][y] == ' ')
		{
			p = crea_punto_schermo(colonna, y, 0);
			pedina  = genera_punto(simbolo_pedina);
			disegna_forma(pedina, p, schermo);
			return 1; 
		}
	}

	return 0;

}

int richiedi_inserimento_pedina()
{
	int colonna;
	printf("In quale colonna vuoi inserire la pedina? ==> ");
	scanf("%d", &colonna);
	return colonna;
}

giocatore crea_giocatore(char nome[MAX_NOME_GIOCATORE], char simbolo_pedina)
{

	giocatore g;
	strcpy(g.nome, nome);
	g.simbolo_pedina = simbolo_pedina;
	return g;
}

void inizializza_schermo(char schermo[SCREEN_W][SCREEN_H])
{

	int x,y;

	for (y = 0; y < SCREEN_H; y++){
		for (x = 0; x < SCREEN_W; x++){
			schermo[x][y] = ' ';
		}
	}	

}

void inserisci_bordi(char schermo[SCREEN_W][SCREEN_H])
{

	int x,y;
	char char_bordo = '%';

	for (y = 0; y < SCREEN_H; y++){
		for (x = 0; x < SCREEN_W; x++){
			if (x == 0 || x == SCREEN_W-1 || y == 0 || y == SCREEN_H-1)
				schermo[x][y] = char_bordo;
		}
	}
}

void disegna_schermo(char schermo[SCREEN_W][SCREEN_H])
{

	pulisci_terminale();
	inserisci_bordi(schermo);

	int x,y;

	for (y = 0; y < SCREEN_H; y++){
		for (x = 0; x < SCREEN_W; x++){
			printf("%c",schermo[x][y]);
		}
		printf("\n");
	}

	
}

forma genera_polinomio(int a2, int a1, int a0, int dim, char carattere)
{


	forma polinomio;
	int x,y,cont = 0;

	for (x = 0; x < dim; x++){
		y = a2 * (x*x) + a1 * x + a0;
		polinomio.pixels[cont].x = x;
		polinomio.pixels[cont].y = y;
		polinomio.pixels[cont].valore = carattere;
		cont++;
	}

	polinomio.numero_pixel = cont;
	polinomio.categoria = F_GENERICA;

	return polinomio;

}

forma genera_quadrato(int dim, char carattere)
{
	forma quadrato;
	int x,y,cont;
	
	cont = 0;

	for (y = 0; y < dim; y++){
		for (x = 0; x<dim; x++){
			quadrato.pixels[cont].x = x;
			quadrato.pixels[cont].y = y;
			quadrato.pixels[cont].valore = carattere;

			cont++;
		} 
	}

	quadrato.numero_pixel = cont;
	quadrato.categoria = F_POLIGONO_QUADRILATERO;

	return quadrato;

}

forma genera_linea(int dim, direzione direzione_linea, char carattere)
{
	forma linea;
	int i,cont;
	
	cont = 0;

	for (i = 0; i<dim; i++){

		if (direzione_linea == D_VERTICALE)
		{
			linea.pixels[cont].x = 0;
			linea.pixels[cont].y = i;
		}
		else if (direzione_linea == D_ORIZZONTALE)
		{
			linea.pixels[cont].x = i;
			linea.pixels[cont].y = 0;
		}
		else
			printf("Errore direzione linea!\n");

		linea.pixels[cont].valore = carattere;

		cont++;
	} 

	linea.numero_pixel = cont;
	linea.categoria = F_LINEA;

	return linea;
}

forma genera_punto(char carattere)
{
	forma punto = genera_linea(1, D_VERTICALE, carattere);
	punto.categoria = F_PUNTO; 

	return punto;
}

void disegna_forma(forma f, punto_schermo p, char schermo[SCREEN_W][SCREEN_H])
{

	int i;
	int x,y;

	for (i = 0; i < f.numero_pixel; i++)
	{
		x = f.pixels[i].x + p.x;
		y = f.pixels[i].y + p.y;

		if (x < SCREEN_W && y < SCREEN_H && x >= 0 && y >= 0)
			schermo[x][y] = f.pixels[i].valore;
	}	

}

punto_schermo crea_punto_schermo(int x, int y, char valore)
{
	punto_schermo p;
	p.x = x;
	p.y = y;
	p.valore = valore;
	return p;
}


void aspetta_invio()
{
	printf("Press enter to continue\n");
	char enter = 0;
	while (enter != '\r' && enter != '\n') { enter = getchar(); }
}

void pulisci_terminale(){

	int i;
	
	for (i = 0; i < LINEE_TERMINALE; i++)
	{
		printf("\n");
	}
}

